"""install: Poetry install."""

from __future__ import annotations

from .utils import ANSI, _doSysExec


def taskInstall(kwargs: list[str]) -> None:
	"""Run the install task.

	Args:
		kwargs (list[str]): additional args
	"""
	_ = kwargs  # unused - silence pylint
	print(f"{ANSI['B']}{ANSI['U']}{ANSI['CB']}Installing{ANSI['CLR']}")
	_doSysExec("poetry install")
	print("Done!")



"""Utils used by the rest of fhutils."""

from __future__ import annotations

import platform
import subprocess
import typing
from pathlib import Path

import tomlkit
import tomlkit.items
from tomlkit import toml_document

PY = "py" if platform.system() == "Windows" else "python3"
ANSI = {
	"B": "\033[01m",
	"CLR": "\033[00m",
	"U": "\033[04m",
	"CB": "\033[36m",
	"CG": "\033[32m",
	"CY": "\033[33m",
	"CR": "\033[31m",
	"CODE": "\033[100m\033[93m",
}


def _getPyproject() -> typing.Any:
	"""Get the pyproject data."""
	with open("pyproject.toml", encoding="utf=8") as pyproject:
		return tomlkit.parse(pyproject.read())


def _setPyproject(toml: toml_document.TOMLDocument):
	"""Write the pyproject data back to file."""
	with open("pyproject.toml", "w", encoding="utf=8") as pyproject:
		pyproject.write(tomlkit.dumps(toml))


def _doSysExec(command: str) -> tuple[int, str]:
	"""Execute a command and check for errors.

	Args:
		command (str): commands as a string

	Raises:
		RuntimeWarning: throw a warning shoANSI['U']d there be a non exit code

	Returns:
		tuple[int, str]: return code + stdout
	"""
	with subprocess.Popen(
		command,
		shell=True,
		stdout=subprocess.PIPE,
		stderr=subprocess.STDOUT,
		encoding="utf-8",
		errors="ignore",
	) as process:
		out = process.communicate()[0]
		exitCode = process.returncode
	return exitCode, out


NAME = str(_getPyproject()["tool"]["poetry"]["name"])



"""audit: dependency checking, security analysis and complexity/ Maintainability...

checking. Use with --fast to speed up the security analysis.
"""

from __future__ import annotations

import json
import math
import os
import subprocess
from sys import exit as sysexit

from .utils import ANSI, NAME, PY, _doSysExec


def getTotalLines() -> int:
	"""Get the total number of lines python files under the project directory.

	Returns:
		int: total number of lines
	"""
	totalLines = 0
	for pydir, _, pyfiles in os.walk(os.getcwd()):
		for pyfile in pyfiles:
			if pyfile.endswith(".py"):
				totalpath = os.path.join(pydir, pyfile)
				totalLines += sum(1 for _i in open(totalpath, "rb"))
	return totalLines


def subtaskScore(totalLines: int) -> None:
	"""CalcANSI['U']ate a score for the modANSI['U']e files using the total ...

	number of lines and pylint output.

	Args:
		totalLines (int): total number of lines for /.
	"""
	# Bugs per KSLOC (J. E. Gaffney, "Estimating the Number of FaANSI['U']ts in Code,"
	# in IEEE Transactions on Software Engineering, vol. SE-10, no. 4, pp.
	# 459-464, JANSI['U']y 1984, doi: 10.1109/TSE.1984.5010260.)
	averageBugsPerLine = 21 / 1000

	# Define pylint args
	pylintArgs = [
		"--enable=all",
		"--disable=pointless-string-statement,superfluous-parens,bad-continuation,wrong-import-position,unsubscriptable-object",
		'--indent-string="\\t"',
		"--ignore-patterns=test_.*?py",
		"--argument-naming-style=camelCase",
		"--attr-naming-style=camelCase",
		"--function-naming-style=camelCase",
		"--method-naming-style=camelCase",
		"--variable-naming-style=camelCase",
		"--load-plugins=pylint.extensions.bad_builtin,pylint.extensions.check_elif,pylint.extensions.redefined_variable_type,pylint.extensions.overlapping_exceptions,pylint.extensions.docparams,pylint.extensions.mccabe,pylint.extensions.empty_comment",
	]
	score = (
		len(
			json.loads(
				_doSysExec(
					"pylint --output-format=json " + " ".join(pylintArgs) + f" {NAME.lower()}"
				)[1]
			)
		)
		/ totalLines
		* 100
	) / averageBugsPerLine
	rank = chr(65 + (score > 49) + (score > 99) + (score > 149) + (score > 199))
	print(
		f"{'Bugs':<26} (%) - {rank} {(score*averageBugsPerLine).__round__(1):>5}\n"
		f"{ANSI['B']}{'Compared to Industry':<26} (%) - {rank} {(score).__round__(1):>5}{ANSI['CLR']}"
	)


def getCCGrade(complexity: float) -> str:
	"""CalcANSI['U']ate the cc grade from the complexity.

	Args:
		complexity (float): the complexity of a file/ project

	Returns:
		str: the grade
	"""
	return chr(65 + min(math.floor(complexity / 10.0) + (5 - complexity < 0), 5))


def subtaskComplexity() -> None:
	"""Report on the complexity of project files."""
	radonCCRaw = json.loads(_doSysExec(PY + " -X utf8 -m radon cc -j  . ")[1])
	ccAve = 0
	for file in radonCCRaw:
		fileStr = file.replace("\\", "/")
		complexity = sum([ccRaw["complexity"] for ccRaw in radonCCRaw[file]]) / len(
			radonCCRaw[file]
		)
		ccAve += complexity
		print(f"{fileStr:<30} - {getCCGrade(complexity)} {complexity.__round__(1):>5}")
	ccAve /= len(radonCCRaw)
	print(f"{ANSI['B']}{'Average':<30} - {getCCGrade(ccAve)} {ccAve.__round__(1):>5}{ANSI['CLR']}")


def subtaskMaintainability() -> None:
	"""Report on the maintainability of project files."""
	radonMIRaw = json.loads(_doSysExec(PY + " -X utf8 -m radon mi -j  . ")[1])
	miAve = 0
	for file in radonMIRaw:
		fileStr = file.replace("\\", "/")
		maintainability = radonMIRaw[file]["mi"]
		miAve += maintainability
		print(f"{fileStr:<30} - {radonMIRaw[file]['rank']} {maintainability.__round__(1):>5}")
	miAve /= len(radonMIRaw)
	rank = chr(65 + (miAve <= 9) + (miAve <= 19))
	print(f"{ANSI['B']}{'Average':<30} - {rank} {miAve.__round__(1):>5}{ANSI['CLR']}")


def subtaskDup(totalLines: int) -> None:
	"""CalcANSI['U']ate the amount of duplicated code using the total number ...

	of lines and pylint output.

	Args:
		totalLines (int): total number of lines
	"""
	pylint = json.loads(
		_doSysExec(
			"pylint --output-format=json --disable=all --enable=duplicate-code --min-similarity-lines 1 "
			+ f" {NAME.lower()}"
		)[1]
	)
	score = sum([message["message"].count("\n") for message in pylint]) / totalLines * 100
	rank = chr(65 + (score > 9) + (score > 19))
	print(f"{ANSI['B']}{'Duplicates':<26} (%) - {rank} {score.__round__(1):>5}{ANSI['CLR']}")


def taskAudit(kwargs: list[str]) -> None:
	"""Do the audit task, this includes checking requirements are up to date...

	security analysis and code complexity metrics

	Args:
		kwargs (list[str]): optional args
	"""
	# Total LOC
	totalLines = getTotalLines()

	# Requirements up to date?
	print(f"{ANSI['B']}{ANSI['U']}{ANSI['CB']}Requirements{ANSI['CLR']}")
	checkrequirements = _doSysExec("checkrequirements -0")
	print(checkrequirements[1])

	# Requirements licenses compatible with project?
	print(f"{ANSI['B']}{ANSI['U']}{ANSI['CB']}Requirement Licenses{ANSI['CLR']}")
	with subprocess.Popen("licensecheck -0") as process:
		licensecheck = (process.wait(),)

	# Do a security analysis
	if "--fast" in kwargs or "--skip" in kwargs:
		cmd = "simplesecurity --fast -0"
	else:
		cmd = "simplesecurity -0"
	simplesecurity = _doSysExec(cmd)
	print(simplesecurity[1])

	# Code score (lower is better)
	print(f"{ANSI['B']}{ANSI['U']}{ANSI['CB']}Score{ANSI['CLR']}")
	subtaskScore(totalLines)

	# Complexity/ Maintainability with radon
	print(f"\n{ANSI['B']}{ANSI['U']}{ANSI['CB']}Complexity{ANSI['CLR']}")
	subtaskComplexity()
	print(f"\n{ANSI['B']}{ANSI['U']}{ANSI['CB']}Maintainability{ANSI['CLR']}")
	subtaskMaintainability()

	# Duplication score (lower is better)
	print(f"\n{ANSI['B']}{ANSI['U']}{ANSI['CB']}Duplication{ANSI['CLR']}")
	subtaskDup(totalLines)

	# Warning for non zero output
	if (checkrequirements[0] | licensecheck[0] | simplesecurity[0]) == 1:
		print(
			f"\n{ANSI['B']}{ANSI['CY']}One of the above checks has reported a warning, "
			f"double check the output of these.{ANSI['CLR']}"
		)
	# Pass non zero to fhmake
	if "-0" in kwargs or "--zero" in kwargs:
		sysexit(checkrequirements[0] | licensecheck[0] | simplesecurity[0])



"""publish: Run poetry publish (interactive)."""

from __future__ import annotations

import subprocess

from .utils import ANSI


def taskPublish(kwargs: list[str]) -> None:
	"""Run the publish task.

	Args:
		kwargs (list[str]): additional args
	"""
	_ = kwargs  # unused - silence pylint
	print(f"{ANSI['B']}{ANSI['U']}{ANSI['CB']}Publish{ANSI['CLR']}")
	with subprocess.Popen("poetry publish") as process:
		_ = process.wait()



"""FredHappyface Makefile for python. Run one of the following subcommands...

install: Poetry install
build: Build documentation, requirements.txt, and run poetry build
publish: Run poetry publish (interactive)
audit: dependency checking, security analysis and complexity/ Maintainability
checking. Use with --fast to speed up the security analysis.
"""
from __future__ import annotations

import argparse

from .build import taskBuild
from .install import taskInstall
from .publish import taskPublish
from .audit import taskAudit


# Add new subcommands here:
COMMAND_MAP = {
	"build": taskBuild,
	"install": taskInstall,
	"publish": taskPublish,
	"audit": taskAudit,
}
HELP = "subcommand must be one of {}".format(list(COMMAND_MAP.keys()))


def cli():
	"""CLI entry point."""
	parser = argparse.ArgumentParser(
		description=__doc__, formatter_class=argparse.RawTextHelpFormatter
	)
	parser.add_argument("subcommand", action="store", help=HELP)
	args, kwargs = parser.parse_known_args()
	if args.subcommand in COMMAND_MAP:
		COMMAND_MAP[args.subcommand](kwargs)
	else:
		print(HELP)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['fhmake']

package_data = \
{'': ['*']}

install_requires = \
['tomlkit>=0.7.0,<2']

extras_require = \
{'full': ['simplesecurity>=2021.2,<2023',
          'checkrequirements>=2021,<2023',
          'poetry>=1.1.2,<3',
          'licensecheck>=2021.1,<2023',
          'fhdoc>=2021,<2023',
          'radon>=4.3.2,<6',
          'pylint>=2.7.2,<3',
          'flake8-polyfill>=1.0.2,<2']}

entry_points = \
{'console_scripts': ['fhmake = fhmake:cli']}

setup_kwargs = {
    'name': 'fhmake',
    'version': '2021.1.1',
    'description': 'FredHappyface Makefile for python. build, install, publish or audit',
    'long_description': '[![GitHub top language](https://img.shields.io/github/languages/top/FHPythonUtils/FHMake.svg?style=for-the-badge)](../../)\n[![Repository size](https://img.shields.io/github/repo-size/FHPythonUtils/FHMake.svg?style=for-the-badge)](../../)\n[![Issues](https://img.shields.io/github/issues/FHPythonUtils/FHMake.svg?style=for-the-badge)](../../issues)\n[![License](https://img.shields.io/github/license/FHPythonUtils/FHMake.svg?style=for-the-badge)](/LICENSE.md)\n[![Commit activity](https://img.shields.io/github/commit-activity/m/FHPythonUtils/FHMake.svg?style=for-the-badge)](../../commits/master)\n[![Last commit](https://img.shields.io/github/last-commit/FHPythonUtils/FHMake.svg?style=for-the-badge)](../../commits/master)\n[![PyPI Downloads](https://img.shields.io/pypi/dm/fhmake.svg?style=for-the-badge)](https://pypistats.org/packages/fhmake)\n[![PyPI Total Downloads](https://img.shields.io/badge/dynamic/json?style=for-the-badge&label=total%20downloads&query=%24.total_downloads&url=https%3A%2F%2Fapi.pepy.tech%2Fapi%2Fprojects%2Ffhmake)](https://pepy.tech/project/fhmake)\n[![PyPI Version](https://img.shields.io/pypi/v/fhmake.svg?style=for-the-badge)](https://pypi.org/project/fhmake)\n\n<!-- omit in toc -->\n# FHMake\n\n<img src="readme-assets/icons/name.png" alt="Project Icon" width="750">\n\n```none\nFredHappyface Makefile for python. Run one of the following subcommands...\n\ninstall: Poetry install\nbuild: Build documentation, requirements.txt, and run poetry build\npublish: Run poetry publish (interactive)\naudit: dependency checking, security analysis and complexity/ Maintainability\nchecking. Use with --fast to speed up the security analysis.\n```\n\n<!-- omit in toc -->\n## Table of Contents\n- [Example Output](#example-output)\n- [Documentation](#documentation)\n- [Install With PIP](#install-with-pip)\n- [Language information](#language-information)\n\t- [Built for](#built-for)\n- [Install Python on Windows](#install-python-on-windows)\n\t- [Chocolatey](#chocolatey)\n\t- [Download](#download)\n- [Install Python on Linux](#install-python-on-linux)\n\t- [Apt](#apt)\n- [How to run](#how-to-run)\n\t- [With VSCode](#with-vscode)\n\t- [From the Terminal](#from-the-terminal)\n- [Download Project](#download-project)\n\t- [Clone](#clone)\n\t\t- [Using The Command Line](#using-the-command-line)\n\t\t- [Using GitHub Desktop](#using-github-desktop)\n\t- [Download Zip File](#download-zip-file)\n- [Community Files](#community-files)\n\t- [Licence](#licence)\n\t- [Changelog](#changelog)\n\t- [Code of Conduct](#code-of-conduct)\n\t- [Contributing](#contributing)\n\t- [Security](#security)\n\t- [Support](#support)\n\t- [Rationale](#rationale)\n\n## Example Output\n\n```none\nfhmake install\n```\n![fhmake install](/readme-assets/terminal/install.svg)\n\n\n```none\nfhmake build\n```\n![fhmake build](/readme-assets/terminal/build.svg)\n\n\n```none\nfhmake publish\n```\n![fhmake publish](/readme-assets/terminal/publish.svg)\n\n\n```none\nfhmake audit\n```\n![fhmake audit](/readme-assets/terminal/audit.svg)\n\n\n## Documentation\nSee the [Docs](/DOCS/) for more information.\n\n## Install With PIP\n\n**"Slim" Build:** Install simplesecurity\\[full\\], poetry, checkrequirements\\[full\\]\nand licensecheck\\[full\\] with pipx\n\n```python\npip install fhmake\n```\n\n**Otherwise:**\n```python\npip install fhmake[full]\n```\n\nHead to https://pypi.org/project/fhmake/ for more info\n\n## Language information\n### Built for\nThis program has been written for Python 3 and has been tested with\nPython version 3.9.0 <https://www.python.org/downloads/release/python-380/>.\n\n## Install Python on Windows\n### Chocolatey\n```powershell\nchoco install python\n```\n### Download\nTo install Python, go to <https://www.python.org/> and download the latest\nversion.\n\n## Install Python on Linux\n### Apt\n```bash\nsudo apt install python3.9\n```\n\n## How to run\n### With VSCode\n1. Open the .py file in vscode\n2. Ensure a python 3.9 interpreter is selected (Ctrl+Shift+P > Python:Select\nInterpreter > Python 3.9)\n3. Run by pressing Ctrl+F5 (if you are prompted to install any modules, accept)\n### From the Terminal\n```bash\n./[file].py\n```\n\n## Download Project\n### Clone\n#### Using The Command Line\n1. Press the Clone or download button in the top right\n2. Copy the URL (link)\n3. Open the command line and change directory to where you wish to\nclone to\n4. Type \'git clone\' followed by URL in step 2\n```bash\n$ git clone https://github.com/FHPythonUtils/FHMake\n```\n\nMore information can be found at\n<https://help.github.com/en/articles/cloning-a-repository>\n\n#### Using GitHub Desktop\n1. Press the Clone or download button in the top right\n2. Click open in desktop\n3. Choose the path for where you want and click Clone\n\nMore information can be found at\n<https://help.github.com/en/desktop/contributing-to-projects/cloning-a-repository-from-github-to-github-desktop>\n\n### Download Zip File\n\n1. Download this GitHub repository\n2. Extract the zip archive\n3. Copy/ move to the desired location\n\n## Community Files\n### Licence\nMIT License\nCopyright (c) FredHappyface\n(See the [LICENSE](/LICENSE.md) for more information.)\n\n### Changelog\nSee the [Changelog](/CHANGELOG.md) for more information.\n\n### Code of Conduct\nOnline communities include people from many backgrounds. The *Project*\ncontributors are committed to providing a friendly, safe and welcoming\nenvironment for all. Please see the\n[Code of Conduct](https://github.com/FHPythonUtils/.github/blob/master/CODE_OF_CONDUCT.md)\n for more information.\n\n### Contributing\nContributions are welcome, please see the\n[Contributing Guidelines](https://github.com/FHPythonUtils/.github/blob/master/CONTRIBUTING.md)\nfor more information.\n\n### Security\nThank you for improving the security of the project, please see the\n[Security Policy](https://github.com/FHPythonUtils/.github/blob/master/SECURITY.md)\nfor more information.\n\n### Support\nThank you for using this project, I hope it is of use to you. Please be aware that\nthose involved with the project often do so for fun along with other commitments\n(such as work, family, etc). Please see the\n[Support Policy](https://github.com/FHPythonUtils/.github/blob/master/SUPPORT.md)\nfor more information.\n\n### Rationale\nThe rationale acts as a guide to various processes regarding projects such as\nthe versioning scheme and the programming styles used. Please see the\n[Rationale](https://github.com/FHPythonUtils/.github/blob/master/RATIONALE.md)\nfor more information.\n',
    'author': 'FredHappyface',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/FHPythonUtils/FHMake',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



"""build: Build documentation, requirements.txt, and run poetry build."""

from __future__ import annotations

import typing

import tomlkit
import tomlkit.items

from .utils import ANSI, _doSysExec, _getPyproject, _setPyproject


def getProcVer(version: str) -> str:
	"""Process a version string. This is pretty opinionated.

	Args:
		version (str): the version

	Returns:
		str: the processed version
	"""
	if version.startswith("^"):
		major = int(version[1:].split(".")[0])
		if major > 1990 or major == 0:  # if cal ver or zero ver
			return f"<{major + 2},>={version[1:]}"
		return f"<{major + 1},>={version[1:]}"
	return version


def getDependencies() -> dict[str, typing.Any]:
	"""Get our dependencies as a dictionary.

	Returns:
		dict[str, str]: [description]
	"""
	return dict(**_getPyproject()["tool"]["poetry"]["dependencies"])


def subtaskGenRequirements() -> None:
	"""Generate the requirements files."""
	dependencies = getDependencies()
	dependencies.pop("python")
	requirements = []
	requirementsOpt = []
	for requirement in dependencies:
		if isinstance(dependencies[requirement], dict):
			dependent = dependencies[requirement]
			if "optional" in dependent and dependent["optional"]:
				requirementsOpt.append(
					f"{requirement}"
					+ f"{'['+dependent['extras'][0]+']' if 'extras' in dependent else ''}"
					f"{getProcVer(dependent['version'])}"
				)
			else:
				requirements.append(
					f"{requirement}"
					+ f"{'['+dependent['extras'][0]+']' if 'extras' in dependent else ''}"
					f"{getProcVer(dependent['version'])}"
				)
		else:
			dependent = typing.cast(str, dependencies[requirement])
			requirements.append(f"{requirement}{getProcVer(dependent)}")
	with open("requirements.txt", "w") as requirementsTxt:
		requirementsTxt.write("\n".join(sorted(requirements)) + "\n")
	with open("requirements_optional.txt", "w") as requirementsTxt:
		requirementsTxt.write("\n".join(sorted(requirements + requirementsOpt)) + "\n")
	print("Done!\n")


def subtaskUpdatePyproject():
	"""Update the pyproject.toml file with our shiny new version specifiers."""
	pyproject = _getPyproject()
	dependencies = pyproject["tool"]["poetry"]["dependencies"]
	for requirement in dependencies:
		if requirement != "python":
			if isinstance(dependencies[requirement], tomlkit.items.InlineTable):
				dependent = dependencies[requirement]["version"]
				dependencies[requirement]["version"] = getProcVer(dependent)
			else:
				dependencies[requirement] = getProcVer(dependencies[requirement])
	_setPyproject(pyproject)


def taskBuild(kwargs: list[str]) -> None:
	"""Run the build task.

	Args:
		kwargs (list[str]): additional args
	"""
	_ = kwargs  # unused - silence pylint
	# Update pyproject.toml version specifiers
	print(f"{ANSI['CG']}(Replacing poetry version specifiers){ANSI['CLR']}")
	subtaskUpdatePyproject()
	# Deal with manual changes to pyproject.toml
	print(f"{ANSI['CG']}(Refreshing Poetry){ANSI['CLR']}\n")
	_doSysExec("poetry update")
	# Generate DOCS
	print(
		f"{ANSI['B']}{ANSI['U']}{ANSI['CB']}Building{ANSI['CLR']}\n\n{ANSI['B']}"
		f"{ANSI['U']}{ANSI['CG']}Documentation{ANSI['CLR']}"
	)
	print(_doSysExec("fhdoc --cleanup")[1].replace("\\", "/"))
	# Generate requirements.txt
	print(f"{ANSI['B']}{ANSI['U']}{ANSI['CG']}Requirements.txt{ANSI['CLR']}")
	subtaskGenRequirements()
	# Generate dist files
	print(f"{ANSI['B']}{ANSI['U']}{ANSI['CG']}Dist files{ANSI['CLR']}")
	print(_doSysExec("poetry build")[1])
